<?php

namespace App\Http\Controllers\Api\V1;

use Auth;
use Illuminate\Http\Request;
use App\Http\Controllers\Controller;
use App\Services\Api\AuthorizeUser;
use App\Quote;
use App\ChatMessage;
use App\Http\Resources\QuoteResource;
use App\Http\Resources\ChatMessageResource;
use App\Http\Resources\FileResource;
use Illuminate\Support\Facades\File;
use Illuminate\Http\UploadedFile;
use Carbon\Carbon;
use Illuminate\Support\Facades\Storage;
use URL;
use OneSignal;

class QuotesController extends Controller
{
    /**
     * This gets a specified quote object.
     *
     * @return Quote Object
     * @author mattsidjohn
     **/
    public function get(Quote $quote)
    {
        return new QuoteResource(
            $quote->with([
                'chatMessages.author',
                'salesManager',
                'salesRep'
            ])->where('id', $quote->id)->get()
        );
    }

    /**
     * Returns all the quote objects.
     *
     * @return Quotes Object
     * @author mattsidjohn
     **/
    public function getAll()
    {
        return new QuoteResource(
            Quote::with([
                'salesManager',
                'salesRep'
            ])->get()
        );
    }

    /**
     * Stores the quote from the request.
     *
     * @return The new quote.
     * @author mattsidjohn
     **/
    public function store(Request $request)
    {
        // return Storage::url('quotes/15/jsyToMI4ev7bayx1DaSafrf9prJNNfJhI25PeNau.jpeg');
        $tmpFilePath = tempnam(sys_get_temp_dir(), $request->fileName);
        file_put_contents($tmpFilePath, base64_decode($request->fileData));
        
        // Use UploadedFile Package
        $uploadedFile =  new UploadedFile(
            $tmpFilePath,
            $request->fileName,
            File::mimeType($tmpFilePath),
            File::size($tmpFilePath),
            null,
            true
        );

        try {
            // Create the new quote with user association
            $quote = Auth::user()
                ->repQuotes()
                ->create(array_merge($request->all(), ['sales_manager_id' => Auth::user()->salesManagers()->first()->id]));
            
            $quote->image = Storage::url(Storage::put('public/quotes/' . $quote->id, $uploadedFile));
            $quote->save();

            return new QuoteResource(
                $quote->with([
                    'salesManager',
                    'salesRep'
                ])->get()
            );
        } catch (Exception $e) {
            return response($e->getMessage, $e->getCode);
        }
    }

    /**
     * Delete the specified resource
     *
     * @return Boolean
     * @author mattsidjohn
     **/
    public function destroy(Quote $quote)
    {
        try {
            // Delete the quote.
            $quote->delete();
            return response('Success!', 200);
        } catch (Exception $e) {
            // Return error if caught
            return response($e->getMessage(), $e->getCode());
        }
    }

    /**
     * Toggle the status of the quote - complete / incomplete
     *
     * @return Boolean
     * @author mattsidjohn
     **/
    public function toggleStatus(Quote $quote)
    {
        try {
            // Update the status of the quote
            $quote->update(['status' => !$quote->status]);

            // Respond with new Quote object
            return new QuoteResource($quote);
        } catch (Exception $e) {
            // Return error if caught
            return response($e->getMessage(), $e->getCode());
        }
    }

    /**
     * Create a new chat message for the quote.
     *
     * @return Message Object
     * @author mattsidjohn
     **/
    public function storeChatMessage(Quote $quote, Request $request)
    {
        try {
            // Create new Chat Message
            $chatMessage = $quote->chatMessages()
                ->create($request->all());

            if (!empty($quote->salesRep) && !empty($quote->salesRep->playerIds)) {
                foreach ($quote->salesRep->playerIds as $playerId) {
                    OneSignal::sendNotificationToUser('There is a new chat message on your quote', $playerId->player_id);
                }
            }

            return new ChatMessageResource(
                $chatMessage->with([
                    'author'
                ])
                ->where('id', $chatMessage->id)
                ->get()
            );
        } catch (Exception $e) {
            return response($e->getMessage, $e->getcode);
        }
    }

    /**
     * Delete the specified resource
     *
     * @return Boolean
     * @author mattsidjohn
     **/
    public function destroyChatMessage(Quote $quote, ChatMessage $chatMessage)
    {
        try {
            // Delete the quote.
            $chatMessage->delete();
            return response('Success!', 200);
        } catch (Exception $e) {
            // Return error if caught
            return response($e->getMessage(), $e->getCode());
        }
    }

    /**
     * Store the given file
     *
     * @return New File Object
     * @author mattsidjohn
     **/
    public function storeFile(Quote $quote, Request $request)
    {
        try {
            // Create new Chat Message
            $file = $quote->files()
                ->create($request->all());
     
            return new FileResource(
                $file->with([
                    'quote'
                ])
                ->where('id', $file->id)
                ->get()
            );
        } catch (Exception $e) {
            return response($e->getMessage, $e->getcode);
        }
    }
}
