<?php

namespace App\Services;

use Illuminate\Http\Request;
use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\Storage;
use App\Competitor;
use App\CompetitorProduct;

class CompetitorService
{
    /**
     * This functions imports and creates or updates competitors in the database.
     *
     * @return Success
     * @author mattsidjohn
     **/
    public function import($file)
    {
        // Convert CSV file to array
        $products = $this->csvToArray($file);

        // Set transaction listener to monitor db transactions
        DB::beginTransaction();

        // Initiate try and catch malformed data insertions resulting in an error
        try {
            $this->createOrUpdate($products);
            
            // Commit transactions to the DB
            DB::commit();
        } catch (Exception $e) {
            // Rollback DB Transaction and return message
            DB::rollback();

            return $e->getMessage();
        }
    }

    /**
     * Create or Update the competitors and their products.
     *
     * @return Success
     * @author mattsidjohn
     **/
    private function createOrUpdate($products)
    {
        // Create or Update each competitor and its products
        foreach ($products as $product) {
            // Check to see if competitor exists.
            $competitor = Competitor::where('name', $product['competitor_id'])
                ->first();

            // If competitor doesn't exist create a new one.
            (empty($competitor)) and $competitor = Competitor::create([
                'name' => $product['competitor_id']
            ]);

            // Add Competitor relationship to the array
            $product['competitor_id'] = $competitor->id;
            
            // Create or Update the competitor product
            CompetitorProduct::updateOrCreate(
                ['item_number' => $product['item_number']],
                $product
            );
        }
    }

    /**
     * This function reads a csv and converts the data to a competitor array
     *
     * @return array
     * @author mattsidjohn
     **/
    public function csvToArray($csv)
    {
        // Open file & create empty array.
        $file = fopen($csv, "r");
        $array = [];

        // Read first line to omit it.
        fgetcsv($file, 10000, ",");

        // Read the contents of the file and pop it to the array.
        while (($row = fgetcsv($file, 0, ",")) !== false) {
            $array[] = [
                'competitor_id' => $row[0],
                'brand_name' => $row[1],
                'material_type' => $row[2],
                'item_number' => $row[3],
                'gallon_capacity' => $row[4],
                'width' => $row[5],
                'length' => $row[6],
                'bag_gauge' => $row[7],
                'strength' => $row[8],
                'color' => $row[9],
                'case_count' => $row[10],
                'rolls_bags' => $row[11],
                'packaging_style' => $row[12],
                'bottom_seal' => $row[13],
                'printed' => $row[14],
                'max_dry_load' => $row[15],
                'skid_count' => $row[16],
                'case_per_layer' => $row[17],
                'case_weight' => $row[18],
                'cube' => $row[19],
                'min_qty' => $row[20],
                'upc' => $row[21]
            ];
        }

        // Close file and return the final array of competitors
        fclose($file);
        return $array;
    }
}
